// existierende Sokoban-Objekte
enum SokobanObjekt {
	Sokoban, Kiste, Wand, Zielfeld
}

// auf einem Feld koennen sich ein oder mehrere SokobanObjekte befinden
class FeldBelegung extends java.util.ArrayList<SokobanObjekt> {
}

class Sokoban {
	private FeldBelegung[][] feld; // aktuelle Belegung aller Felder

	// Voraussetzung: die Datei enthaelt eine korrekte ASCII-Repraesentation
	// eines Sokoban-Spielfeldes
	public Sokoban(String dateiname) {
		char[][] feld = IO.readFileAsCharMatrix(dateiname);
		this.feld = convert(feld);
	}

	public void spielen() {
		this.print();
		while (!spielEnde()) {
			char eingabe = IO.readChar("N/O/W/S?");
			switch (eingabe) {
			case 'N':
			case 'n':
				nachNorden();
				break;
			case 'O':
			case 'o':
				nachOsten();
				break;
			case 'W':
			case 'w':
				nachWesten();
				break;
			case 'S':
			case 's':
				nachSueden();
				break;
			}
			this.print();
		}
		System.out.println("Geschafft!");
	}

	void print() {
		for (int r = 0; r < feld.length; r++) {
			for (int s = 0; s < feld[0].length; s++) {
				if (feld[r][s].isEmpty()) {
					System.out.print(' ');
				} else if (feld[r][s].contains(SokobanObjekt.Wand)) {
					System.out.print('#');
				} else if (feld[r][s].size() == 1) {
					if (feld[r][s].contains(SokobanObjekt.Sokoban)) {
						System.out.print('@');
					} else if (feld[r][s].contains(SokobanObjekt.Kiste)) {
						System.out.print('$');
					} else { // Zielfeld
						System.out.print('.');
					}
				} else { // zwei Objekte
					if (feld[r][s].contains(SokobanObjekt.Sokoban)
							&& feld[r][s].contains(SokobanObjekt.Zielfeld)) {
						System.out.print('+');
					} else { // Kiste und Zielfeld
						System.out.print('*');
					}
				}
			}
			System.out.println();
		}
		System.out.println();
		System.out.println();
	}

	boolean spielEnde() {
		// Spielende ist erreich, wenn alle Kisten auf einem Zielfeld stehen
		for (int r = 0; r < feld.length; r++) {
			for (int s = 0; s < feld[0].length; s++) {
				if (feld[r][s].contains(SokobanObjekt.Kiste)) {
					if (!feld[r][s].contains(SokobanObjekt.Zielfeld)) {
						return false;
					}
				}
			}
		}
		return true;
	}

	// konvertiert ein Sokoban-Spielfeld aus dem ASCII-Format in eine interne
	// Repraesentation
	// Achtung: Es wird vorausgesetzt, dass die ASCII-Datei korrekt ist!
	static FeldBelegung[][] convert(char[][] feld) {
		FeldBelegung[][] sokobanFeld = new FeldBelegung[feld.length][feld[0].length];
		for (int r = 0; r < feld.length; r++) {
			for (int s = 0; s < feld[0].length; s++) {
				sokobanFeld[r][s] = new FeldBelegung();
				switch (feld[r][s]) {
				case '#':
					sokobanFeld[r][s].add(SokobanObjekt.Wand);
					break;
				case '@':
					sokobanFeld[r][s].add(SokobanObjekt.Sokoban);
					break;
				case '.':
					sokobanFeld[r][s].add(SokobanObjekt.Zielfeld);
					break;
				case '$':
					sokobanFeld[r][s].add(SokobanObjekt.Kiste);
					break;
				case '*':
					sokobanFeld[r][s].add(SokobanObjekt.Kiste);
					sokobanFeld[r][s].add(SokobanObjekt.Zielfeld);
					break;
				case '+':
					sokobanFeld[r][s].add(SokobanObjekt.Sokoban);
					sokobanFeld[r][s].add(SokobanObjekt.Zielfeld);
					break;
				}
			}
		}
		return sokobanFeld;
	}

	void nachNorden() {
		Position sokobanPos = findeSokoban();
		if (feld[sokobanPos.reihe - 1][sokobanPos.spalte]
				.contains(SokobanObjekt.Wand)) {
			return; // oberhalb befindet sich eine Wand
		}
		if (!feld[sokobanPos.reihe - 1][sokobanPos.spalte]
				.contains(SokobanObjekt.Kiste)) {
			// Oberhalb befindet sich kein Hindernis, d.h. Sokoban kann nach
			// oben verschoben werden
			feld[sokobanPos.reihe][sokobanPos.spalte]
					.remove(SokobanObjekt.Sokoban);
			feld[sokobanPos.reihe - 1][sokobanPos.spalte]
					.add(SokobanObjekt.Sokoban);
			return;
		}

		// Oberhalb befindet sich eine Kiste
		// ueberpruefe, ob Hindernis oberhalb der Kiste vorhanden ist
		if (feld[sokobanPos.reihe - 2][sokobanPos.spalte]
				.contains(SokobanObjekt.Kiste)
				|| feld[sokobanPos.reihe - 2][sokobanPos.spalte]
						.contains(SokobanObjekt.Wand)) {
			return; // Kiste kann nicht verschoben werden
		}

		// Kiste und Sokoban werden verschoben
		feld[sokobanPos.reihe - 1][sokobanPos.spalte]
				.remove(SokobanObjekt.Kiste);
		feld[sokobanPos.reihe - 2][sokobanPos.spalte].add(SokobanObjekt.Kiste);
		feld[sokobanPos.reihe][sokobanPos.spalte].remove(SokobanObjekt.Sokoban);
		feld[sokobanPos.reihe - 1][sokobanPos.spalte]
				.add(SokobanObjekt.Sokoban);

	}

	void nachSueden() {
		Position sokobanPos = findeSokoban();
		if (feld[sokobanPos.reihe + 1][sokobanPos.spalte]
				.contains(SokobanObjekt.Wand)) {
			return; // unterhalb befindet sich eine Wand
		}
		if (!feld[sokobanPos.reihe + 1][sokobanPos.spalte]
				.contains(SokobanObjekt.Kiste)) {
			// unterhalb befindet sich kein Hindernis, d.h. Sokoban kann nach
			// oben verschoben werden
			feld[sokobanPos.reihe][sokobanPos.spalte]
					.remove(SokobanObjekt.Sokoban);
			feld[sokobanPos.reihe + 1][sokobanPos.spalte]
					.add(SokobanObjekt.Sokoban);
			return;
		}

		// Unterhalb befindet sich eine Kiste
		// ueberpruefe, ob Hindernis unterhalb der Kiste vorhanden ist
		if (feld[sokobanPos.reihe + 2][sokobanPos.spalte]
				.contains(SokobanObjekt.Kiste)
				|| feld[sokobanPos.reihe + 2][sokobanPos.spalte]
						.contains(SokobanObjekt.Wand)) {
			return; // Kiste kann nicht verschoben werden
		}

		// Kiste und Sokoban werden verschoben
		feld[sokobanPos.reihe + 1][sokobanPos.spalte]
				.remove(SokobanObjekt.Kiste);
		feld[sokobanPos.reihe + 2][sokobanPos.spalte].add(SokobanObjekt.Kiste);
		feld[sokobanPos.reihe][sokobanPos.spalte].remove(SokobanObjekt.Sokoban);
		feld[sokobanPos.reihe + 1][sokobanPos.spalte]
				.add(SokobanObjekt.Sokoban);

	}

	void nachWesten() {
		Position sokobanPos = findeSokoban();
		if (feld[sokobanPos.reihe][sokobanPos.spalte - 1]
				.contains(SokobanObjekt.Wand)) {
			return; // links befindet sich eine Wand
		}
		if (!feld[sokobanPos.reihe][sokobanPos.spalte - 1]
				.contains(SokobanObjekt.Kiste)) {
			// links befindet sich kein Hindernis, d.h. Sokoban kann nach
			// links verschoben werden
			feld[sokobanPos.reihe][sokobanPos.spalte]
					.remove(SokobanObjekt.Sokoban);
			feld[sokobanPos.reihe][sokobanPos.spalte - 1]
					.add(SokobanObjekt.Sokoban);
			return;
		}

		// Links befindet sich eine Kiste
		// ueberpruefe, ob Hindernis links der Kiste vorhanden ist
		if (feld[sokobanPos.reihe][sokobanPos.spalte - 2]
				.contains(SokobanObjekt.Kiste)
				|| feld[sokobanPos.reihe][sokobanPos.spalte - 2]
						.contains(SokobanObjekt.Wand)) {
			return; // Kiste kann nicht verschoben werden
		}

		// Kiste und Sokoban werden verschoben
		feld[sokobanPos.reihe][sokobanPos.spalte - 1]
				.remove(SokobanObjekt.Kiste);
		feld[sokobanPos.reihe][sokobanPos.spalte - 2].add(SokobanObjekt.Kiste);
		feld[sokobanPos.reihe][sokobanPos.spalte].remove(SokobanObjekt.Sokoban);
		feld[sokobanPos.reihe][sokobanPos.spalte - 1]
				.add(SokobanObjekt.Sokoban);

	}

	void nachOsten() {
		Position sokobanPos = findeSokoban();
		if (feld[sokobanPos.reihe][sokobanPos.spalte + 1]
				.contains(SokobanObjekt.Wand)) {
			return; // rechts befindet sich eine Wand
		}
		if (!feld[sokobanPos.reihe][sokobanPos.spalte + 1]
				.contains(SokobanObjekt.Kiste)) {
			// rechts befindet sich kein Hindernis, d.h. Sokoban kann nach
			// rechts verschoben werden
			feld[sokobanPos.reihe][sokobanPos.spalte]
					.remove(SokobanObjekt.Sokoban);
			feld[sokobanPos.reihe][sokobanPos.spalte + 1]
					.add(SokobanObjekt.Sokoban);
			return;
		}

		// Rechts befindet sich eine Kiste
		// ueberpruefe, ob Hindernis rechts der Kiste vorhanden ist
		if (feld[sokobanPos.reihe][sokobanPos.spalte + 2]
				.contains(SokobanObjekt.Kiste)
				|| feld[sokobanPos.reihe][sokobanPos.spalte + 2]
						.contains(SokobanObjekt.Wand)) {
			return; // Kiste kann nicht verschoben werden
		}

		// Kiste und Sokoban werden verschoben
		feld[sokobanPos.reihe][sokobanPos.spalte + 1]
				.remove(SokobanObjekt.Kiste);
		feld[sokobanPos.reihe][sokobanPos.spalte + 2].add(SokobanObjekt.Kiste);
		feld[sokobanPos.reihe][sokobanPos.spalte].remove(SokobanObjekt.Sokoban);
		feld[sokobanPos.reihe][sokobanPos.spalte + 1]
				.add(SokobanObjekt.Sokoban);

	}

	Position findeSokoban() {
		for (int r = 0; r < feld.length; r++) {
			for (int s = 0; s < feld[0].length; s++) {
				if (feld[r][s].contains(SokobanObjekt.Sokoban)) {
					return new Position(r, s);
				}
			}
		}
		return null; // kann nicht sein
	}
}

public class UE16Aufgabe12 {

	public static void main(String[] args) {
		Sokoban sokoban = new Sokoban("bsp2.soko");
		sokoban.spielen();
	}
}

class Position {
	int reihe;
	int spalte;

	Position(int r, int s) {
		reihe = r;
		spalte = s;
	}
}
